# -*- coding: utf-8 -*-

import requests
import filecmp
import calendar
import json
import base64

from oss2.exceptions import (ClientError, RequestError, NoSuchBucket, OpenApiServerError,
        NotFound, NoSuchKey, Conflict, PositionNotEqualToLength, ObjectNotAppendable)

from oss2.compat import is_py2, is_py33
from oss2.models import Tagging, TaggingRule
from oss2.headers import OSS_OBJECT_TAGGING, OSS_OBJECT_TAGGING_COPY_DIRECTIVE
from oss2.compat import urlunquote, urlquote, to_bytes, to_string

from .common import *


def now():
    return int(calendar.timegm(time.gmtime()))


class TestObject(OssTestCase):
    def test_object(self):
        key = self.random_key('.js')
        content = random_bytes(1024)

        self.assertRaises(NotFound, self.bucket.head_object, key)

        lower_bound = now() - 60 * 16
        upper_bound = now() + 60 * 16

        def assert_result(result):
            self.assertEqual(result.content_length, len(content))
            self.assertEqual(result.content_type, 'application/javascript')
            self.assertEqual(result.object_type, 'Normal')

            self.assertTrue(result.last_modified > lower_bound)
            self.assertTrue(result.last_modified < upper_bound)

            self.assertTrue(result.etag)

        self.bucket.put_object(key, content)

        get_result = self.bucket.get_object(key)
        self.assertEqual(get_result.read(), content)
        assert_result(get_result)
        self.assertTrue(get_result.client_crc is not None)
        self.assertTrue(get_result.server_crc is not None)
        self.assertTrue(get_result.client_crc == get_result.server_crc)

        head_result = self.bucket.head_object(key)
        assert_result(head_result)

        self.assertEqual(get_result.last_modified, head_result.last_modified)
        self.assertEqual(get_result.etag, head_result.etag)

        self.bucket.delete_object(key)

        self.assertRaises(NoSuchKey, self.bucket.get_object, key)

    def test_restore_object(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-restore-object"
        bucket = oss2.Bucket(auth, OSS_ENDPOINT, bucket_name)

        bucket.create_bucket(oss2.BUCKET_ACL_PRIVATE, oss2.models.BucketCreateConfig(oss2.BUCKET_STORAGE_CLASS_ARCHIVE))

        service = oss2.Service(auth, OSS_ENDPOINT)
        wait_meta_sync()
        self.retry_assert(lambda: bucket.bucket_name in (b.name for b in
                                                         service.list_buckets(prefix=bucket.bucket_name).buckets))

        key = 'a.txt'
        bucket.put_object(key, 'content')
        self.assertEqual(202, bucket.restore_object(key).status)
        bucket.delete_object(key)
        bucket.delete_bucket()

    def test_restore_object_with_config(self):
        from oss2.models import (RestoreJobParameters, RestoreConfiguration, RESTORE_TIER_EXPEDITED,
                                 RESTORE_TIER_STANDARD, RESTORE_TIER_BULK)

        endpoint = "http://oss-ap-southeast-2.aliyuncs.com"
        bucket_name = OSS_BUCKET + "-test-restore-1"
        bucket = oss2.Bucket(oss2.make_auth(OSS_ID, OSS_SECRET, OSS_AUTH_VERSION), endpoint, bucket_name)
        bucket.create_bucket()

        prefix = "test-restore-object-with-request"
        tiers = [RESTORE_TIER_EXPEDITED, RESTORE_TIER_STANDARD, RESTORE_TIER_BULK]

        for index in range(0, len(tiers)):
            object_name = prefix + str(index) + ".txt"
            bucket.put_object(object_name, '123', headers={"x-oss-storage-class": oss2.BUCKET_STORAGE_CLASS_COLD_ARCHIVE})

            meta = bucket.head_object(object_name)
            self.assertEqual(oss2.BUCKET_STORAGE_CLASS_COLD_ARCHIVE, meta.resp.headers['x-oss-storage-class'])

            job_parameters = RestoreJobParameters(tiers[index])
            restore_config= oss2.models.RestoreConfiguration(days=5, job_parameters=job_parameters)
            result = bucket.restore_object(object_name, input=restore_config)
            self.assertEqual(tiers[index], result.resp.headers.get('x-oss-object-restore-priority'))

    def test_restore_object_with_wrong_request(self):
        from oss2.models import RestoreJobParameters, RestoreConfiguration

        endpoint = "http://oss-ap-southeast-2.aliyuncs.com"
        bucket_name = OSS_BUCKET + "-test-restore-2"
        bucket = oss2.Bucket(oss2.make_auth(OSS_ID, OSS_SECRET, OSS_AUTH_VERSION), endpoint, bucket_name)
        bucket.create_bucket()

        object_name = "test_restore_object_with_wrong_configuration.txt"
        bucket.put_object(object_name, '123', headers={"x-oss-storage-class": oss2.BUCKET_STORAGE_CLASS_COLD_ARCHIVE})
        meta = bucket.head_object(object_name)
        self.assertEqual(oss2.BUCKET_STORAGE_CLASS_COLD_ARCHIVE, meta.resp.headers['x-oss-storage-class'])

        # wrong tier
        job_parameters = RestoreJobParameters('wrong-tier')
        restore_config = oss2.models.RestoreConfiguration(days=5, job_parameters=job_parameters)
        self.assertRaises(oss2.exceptions.InvalidArgument, bucket.restore_object, object_name, input=restore_config)

    def test_restore_archive_object_with_job_parameters(self):
        from oss2.models import RestoreJobParameters, RestoreConfiguration, RESTORE_TIER_BULK

        endpoint = "http://oss-ap-southeast-2.aliyuncs.com"
        bucket_name = OSS_BUCKET + "-test-restore-3"
        bucket = oss2.Bucket(oss2.make_auth(OSS_ID, OSS_SECRET, OSS_AUTH_VERSION), endpoint, bucket_name)
        bucket.create_bucket()

        object_name = "test_restore_archive_object_with_job_parameters.txt"
        bucket.put_object(object_name, '123', headers={"x-oss-storage-class": oss2.BUCKET_STORAGE_CLASS_ARCHIVE})
        meta = bucket.head_object(object_name)
        self.assertEqual(oss2.BUCKET_STORAGE_CLASS_ARCHIVE, meta.resp.headers['x-oss-storage-class'])

        job_parameters = RestoreJobParameters(RESTORE_TIER_BULK)
        restore_config = oss2.models.RestoreConfiguration(days=5, job_parameters=job_parameters)
        self.assertRaises(oss2.exceptions.MalformedXml, bucket.restore_object, object_name, input=restore_config)

    def test_last_modified_time(self):
        key = self.random_key()
        content = random_bytes(10)

        self.bucket.put_object(key, content)

        res = self.bucket.get_object(key)
        res.read()

        time_string = res.headers['Last-Modified']
        self.assertEqual(oss2.utils.http_to_unixtime(time_string), res.last_modified)
        self.assertEqual(oss2.utils.to_unixtime(time_string, '%a, %d %b %Y %H:%M:%S GMT'), res.last_modified)

    def test_file(self):
        filename = random_string(12) + '.js'
        filename2 = random_string(12)

        key = self.random_key('.txt')
        content = random_bytes(1024 * 1024)

        with open(filename, 'wb') as f:
            f.write(content)

        # 上传本地文件到OSS
        self.bucket.put_object_from_file(key, filename)

        # 检查Content-Type应该是javascript
        result = self.bucket.head_object(key)
        self.assertEqual(result.headers['content-type'], 'application/javascript')

        # 下载到本地文件
        get_result = self.bucket.get_object_to_file(key, filename2)

        self.assertTrue(filecmp.cmp(filename, filename2))

        # 上传本地文件的一部分到OSS
        key_partial = self.random_key('-partial.txt')
        offset = 100
        with open(filename, 'rb') as f:
            f.seek(offset, os.SEEK_SET)
            self.bucket.put_object(key_partial, f)

        # 检查上传后的文件
        result = self.bucket.get_object(key_partial)
        self.assertEqual(result.content_length, len(content) - offset)
        self.assertEqual(result.read(), content[offset:])

        # 清理
        os.remove(filename)
        os.remove(filename2)

    def test_object_empty(self):
        key = self.random_key()
        content = b''

        self.bucket.put_object(key, content)
        res = self.bucket.get_object(key)

        self.assertEqual(res.server_crc, 0)

    def test_file_empty(self):
        input_filename = random_string(12)
        output_filename = random_string(12)

        key = self.random_key()
        content = b''

        with open(input_filename, 'wb') as f:
            f.write(content)

        self.bucket.put_object_from_file(key, input_filename)
        self.bucket.get_object_to_file(key, output_filename)

        self.assertTrue(filecmp.cmp(input_filename, output_filename))

        os.remove(input_filename)
        os.remove(output_filename)

    def test_streaming(self):
        src_key = self.random_key('.src')
        dst_key = self.random_key('.dst')

        content = random_bytes(1024 * 1024)

        self.bucket.put_object(src_key, content)

        # 获取OSS上的文件，一边读取一边写入到另外一个OSS文件
        src = self.bucket.get_object(src_key)
        result = self.bucket.put_object(dst_key, src)

        # verify        
        self.assertTrue(src.client_crc is not None)
        self.assertTrue(src.server_crc is not None)  
        self.assertEqual(src.client_crc, src.server_crc)
        self.assertEqual(result.crc, src.server_crc)
        self.assertEqual(self.bucket.get_object(src_key).read(), self.bucket.get_object(dst_key).read())

    def make_generator(self, content, chunk_size):
        def generator():
            offset = 0
            while offset < len(content):
                n = min(chunk_size, len(content) - offset)
                yield content[offset:offset+n]

                offset += n

        return generator()

    def test_data_generator(self):
        key = self.random_key()
        key2 = self.random_key()
        content = random_bytes(1024 * 1024 + 1)

        self.bucket.put_object(key, self.make_generator(content, 8192))
        self.assertEqual(self.bucket.get_object(key).read(), content)

        # test progress
        stats = {'previous': -1}

        def progress_callback(bytes_consumed, total_bytes):
            self.assertTrue(total_bytes is None)
            self.assertTrue(bytes_consumed > stats['previous'])

            stats['previous'] = bytes_consumed

        self.bucket.put_object(key2, self.make_generator(content, 8192), progress_callback=progress_callback)
        self.assertEqual(self.bucket.get_object(key).read(), content)

    def test_request_error(self):
        bad_endpoint = random_string(8) + '.' + random_string(16) + '.com'
        bucket = oss2.Bucket(oss2.Auth(OSS_ID, OSS_SECRET), bad_endpoint, OSS_BUCKET)

        try:
            bucket.get_bucket_acl()
        except RequestError as e:
            self.assertEqual(e.status, oss2.exceptions.OSS_REQUEST_ERROR_STATUS)
            self.assertEqual(e.request_id, '')
            self.assertEqual(e.code, '')
            self.assertEqual(e.message, '')

            self.assertTrue(str(e))
            self.assertTrue(e.body)

    def test_timeout(self):
        bucket = oss2.Bucket(oss2.Auth(OSS_ID, OSS_SECRET), OSS_ENDPOINT, OSS_BUCKET,
                             connect_timeout=0.001)
        self.assertRaises(RequestError, bucket.get_bucket_acl)

    def test_default_timeout(self):
        oss2.defaults.connect_timeout = 0.001
        bucket = oss2.Bucket(oss2.Auth(OSS_ID, OSS_SECRET), OSS_ENDPOINT, OSS_BUCKET)
        self.assertRaises(RequestError, bucket.get_bucket_acl)
        oss2.defaults.connect_timeout = 60

    def test_get_object_iterator(self):
        key = self.random_key()
        content = random_bytes(1024 * 1024)

        self.bucket.put_object(key, content)
        result = self.bucket.get_object(key)
        content_got = b''

        for chunk in result:
            content_got += oss2.to_bytes(chunk)

        self.assertEqual(len(content), len(content_got))
        self.assertEqual(content, content_got)

        result = self.bucket.get_object(key)
        content_got = b''

        for chunk in result:
            content_got += chunk

        self.assertEqual(len(content), len(content_got))
        self.assertEqual(content, content_got)

    def test_query_parameter(self):
        key = self.random_key()
        content = random_bytes(1024 * 1024)
        self.bucket.put_object(key, content, headers={'Content-Type': 'plain/text'})
        query_params = {'response-content-type': 'image/jpeg'}
        result = self.bucket.get_object(key, params=query_params)
        self.assertEqual(result.headers['content-type'], 'image/jpeg')

    def test_anonymous(self):
        key = self.random_key()
        content = random_bytes(512)

        # 设置bucket为public-read，并确认可以上传和下载
        self.bucket.put_bucket_acl('public-read-write')
        wait_meta_sync()

        b = oss2.Bucket(oss2.AnonymousAuth(), OSS_ENDPOINT, OSS_BUCKET)
        b.put_object(key, content)
        result = b.get_object(key)
        self.assertEqual(result.read(), content)

        # 测试sign_url
        url = b.sign_url('GET', key, 100, params={'para1':'test'})
        resp = requests.get(url)
        self.assertEqual(content, resp.content)

        # 设置bucket为private，并确认上传和下载都会失败
        self.bucket.put_bucket_acl('private')
        wait_meta_sync()

        self.assertRaises(oss2.exceptions.AccessDenied, b.put_object, key, content)
        self.assertRaises(oss2.exceptions.AccessDenied, b.get_object, key)

    def test_range_get(self):
        key = self.random_key()
        content = random_bytes(1024)

        self.bucket.put_object(key, content)

        result = self.bucket.get_object(key, byte_range=(500, None))
        self.assertEqual(result.read(), content[500:])

        result = self.bucket.get_object(key, byte_range=(None, 199))
        self.assertEqual(result.read(), content[-199:])

        result = self.bucket.get_object(key, byte_range=(3, 3))
        self.assertEqual(result.read(), content[3:4])

    def test_list_objects(self):
        object_name = 'test-list-objects'
        self.bucket.put_object(object_name, b'a')

        result = self.bucket.list_objects()
        self.assertEqual(result.status, 200)
        self.assertTrue(len(result.object_list) > 0)
        for obj in result.object_list:
            self.assertIsNotNone(obj.owner)
            self.assertTrue(len(obj.owner.id) > 0)
            self.assertTrue(len(obj.owner.display_name) > 0)

    def test_batch_delete_objects(self):
        object_list = []
        for i in range(0, 5):
            key = self.random_key()
            object_list.append(key)

            self.bucket.put_object(key, random_string(64))

        result = self.bucket.batch_delete_objects(object_list)
        self.assertEqual(sorted(object_list), sorted(result.deleted_keys))
        keys = []
        for i in range(0, 5):
            keys.append(result.delete_versions[i].key)
        self.assertEqual(sorted(object_list), sorted(keys))
        self.assertEqual(5, len(result.deleted_keys))
        self.assertEqual(5, len(result.delete_versions))

        for object in object_list:
            self.assertTrue(not self.bucket.object_exists(object))

    def test_batch_delete_objects_empty(self):
        try:
            self.bucket.batch_delete_objects([])
        except ClientError as e:
            self.assertEqual(e.status, oss2.exceptions.OSS_CLIENT_ERROR_STATUS)
            self.assertEqual(e.request_id, '')
            self.assertEqual(e.code, '')
            self.assertEqual(e.message, '')

            self.assertTrue(e.body)
            self.assertTrue(str(e))

    def test_append_object(self):
        key = self.random_key()
        content1 = random_bytes(512)
        content2 = random_bytes(128)

        result = self.bucket.append_object(key, 0, content1, init_crc=0)
        self.assertEqual(result.next_position, len(content1))
        self.assertTrue(result.crc is not None)

        try:
            self.bucket.append_object(key, 0, content2)
        except PositionNotEqualToLength as e:
            self.assertEqual(e.next_position, len(content1))
        else:
            self.assertTrue(False)
        
        result = self.bucket.append_object(key, len(content1), content2, init_crc=result.crc)
        self.assertEqual(result.next_position, len(content1) + len(content2))
        self.assertTrue(result.crc is not None)

        self.bucket.delete_object(key)

    def test_private_download_url(self):
        for key in [self.random_key(), self.random_key(u'中文文件名')]:
            content = random_bytes(42)

            self.bucket.put_object(key, content)
            url = self.bucket.sign_url('GET', key, 60)

            resp = requests.get(url)
            self.assertEqual(content, resp.content)
            
    def test_sign_url_with_callback(self):
        key = self.random_key()
        
        def encode_callback(cb_dict):
            cb_str = json.dumps(callback_params).strip()
            return oss2.compat.to_string(base64.b64encode(oss2.compat.to_bytes(cb_str))) 
        
        # callback
        callback_params = {}
        callback_params['callbackUrl'] = 'http://cbsrv.oss.demo.com'
        callback_params['callbackBody'] = 'bucket=${bucket}&object=${object}' 
        encoded_callback = encode_callback(callback_params)
        
        # callback vars
        callback_var_params = {'x:my_var1': 'my_val1', 'x:my_var2': 'my_val2'}
        encoded_callback_var = encode_callback(callback_var_params)
        
        # put with callback
        params = {'callback': encoded_callback, 'callback-var': encoded_callback_var}
        url = self.bucket.sign_url('PUT', key, 60, params=params)
        resp = requests.put(url)
        self.assertEqual(resp.status_code, 203)

    def test_private_download_url_with_extra_query(self):
        if os.getenv('OSS_TEST_AUTH_VERSION') != oss2.AUTH_VERSION_2:
            return
        key = self.random_key()
        content = random_bytes(42)

        self.bucket.put_object(key, content)
        url = self.bucket.sign_url('GET', key, 60, params={'extra-query': '1'})

        resp = requests.get(url)
        self.assertEqual(content, resp.content)

        resp = requests.get(url + '&another-query=1')
        self.assertEqual(resp.status_code, 403)

        e = oss2.exceptions.make_exception(oss2.http.Response(resp))
        self.assertEqual(e.status, 403)
        self.assertEqual(e.code, 'SignatureDoesNotMatch')

    def test_put_object_with_sign_url(self):
        key = self.random_key(".jpg")
        with open("tests/example.jpg", 'rb') as fr:
            data = fr.read()

        url = self.bucket.sign_url('PUT', key, 3600)
        result = self.bucket.put_object_with_url(url, data)
        self.assertEqual(result.status, 200)
        result = self.bucket.head_object(key)
        self.assertEqual(result.content_type, "application/octet-stream")

        headers = {'Content-Type': "image/jpeg"}
        self.assertRaises(oss2.exceptions.SignatureDoesNotMatch, self.bucket.put_object_with_url, url, data, headers=headers)

        url = self.bucket.sign_url('PUT', key, 3600, headers=headers)
        self.assertRaises(oss2.exceptions.SignatureDoesNotMatch, self.bucket.put_object_with_url, url, data)
        result = self.bucket.put_object_with_url(url, data, headers)
        self.assertEqual(result.status, 200)
        result = self.bucket.head_object(key)
        self.assertEqual(result.content_type, "image/jpeg")

    def test_get_object_with_sign_url(self):
        key = self.random_key('.txt')
        content = random_bytes(100)

        result = self.bucket.put_object(key, content)
        self.assertEqual(result.status, 200)

        # normal get with signed url
        url = self.bucket.sign_url("GET", key, 3600)
        result = self.bucket.get_object_with_url(url)
        self.assertEqual(result.status, 200)

        # signed without range, and get with range
        result = self.bucket.get_object_with_url(url, byte_range=(50, 99))
        self.assertEqual(result.status, 206)
        range_content = result.read()
        self.assertEqual(content[50:], range_content)

        # signed with range, and get without range
        headers = {'Range': 'bytes=50-99'}
        url = self.bucket.sign_url("GET", key, 3600, headers=headers)
        result = self.bucket.get_object_with_url(url)
        self.assertEqual(result.status, 200)
        range_content = result.read()
        self.assertEqual(content, range_content)

    def test_put_object_from_file_with_sign_url(self):
        key = self.random_key()
        file_name = self.random_filename()
        content = random_bytes(100)

        with open(file_name, 'wb') as fw:
            fw.write(content)

        headers = {'Content-Type': "text/plain"}
        url = self.bucket.sign_url('PUT', key, 3600, headers)
        result = self.bucket.put_object_with_url_from_file(url, file_name, headers=headers)
        self.assertEqual(result.status, 200)
        result = self.bucket.head_object(key)
        self.assertEqual(result.content_type, "text/plain")

    def test_get_object_to_file_with_sign_url(self):
        key = self.random_key('txt')
        file_name = self.random_filename()
        content = random_bytes(100)

        result = self.bucket.put_object(key, content)
        self.assertEqual(result.status, 200)

        url = self.bucket.sign_url("GET", key, 3600)
        result = self.bucket.get_object_with_url_to_file(url, file_name)
        self.assertEqual(result.status, 200)

    def test_get_object_with_sign_url_slash_safe(self):
        key = 'url下载测试斜杠保护+/二层目录+/测+试斜杠保护.object'
        slash_safe_key = urlquote('url下载测试斜杠保护+') + '/' + urlquote('二层目录+') + '/' + urlquote('测+试斜杠保护.object')
        content = random_bytes(100)

        result = self.bucket.put_object(key, content)
        self.assertEqual(result.status, 200)

        # 不带slash_safe参数
        url = self.bucket.sign_url("GET", key, 60)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertEqual(seek, -1)

        result = self.bucket.get_object_with_url(url)
        self.assertEqual(result.status, 200)

        # slash_safe = False
        url = self.bucket.sign_url("GET", key, 60, slash_safe=False)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertEqual(seek, -1)

        result = self.bucket.get_object_with_url(url)
        self.assertEqual(result.status, 200)

        # slash_safe = True
        url = self.bucket.sign_url("GET", key, 60, slash_safe=True)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertTrue(seek != -1)

        result = self.bucket.get_object_with_url(url)
        self.assertEqual(result.status, 200)

        self.bucket.delete_object(key)
    
    def test_put_object_with_sign_url_slash_safe(self):
        key = 'url上传测试斜杠保护+/二层目录+/测+试斜杠保护.object'
        slash_safe_key = urlquote('url上传测试斜杠保护+') + '/' + urlquote('二层目录+') + '/' + urlquote('测+试斜杠保护.object')
        content = random_bytes(1024)

        # 不带slash_safe 参数
        url = self.bucket.sign_url('PUT', key, 60)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertEqual(seek, -1)

        result = self.bucket.put_object_with_url(url, content)
        self.assertEqual(result.status, 200)

        result = self.bucket.head_object(key)
        self.assertEqual(result.content_length, 1024)

        result = self.bucket.delete_object(key)
        self.assertEqual(result.status, 204)

        # slash_safe = False
        url = self.bucket.sign_url('PUT', key, 60, slash_safe=False)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertEqual(seek, -1)

        result = self.bucket.put_object_with_url(url, content)
        self.assertEqual(result.status, 200)

        result = self.bucket.head_object(key)
        self.assertEqual(result.content_length, 1024)
        
        result = self.bucket.delete_object(key)
        self.assertEqual(result.status, 204)

        # slash_safe = True
        url = self.bucket.sign_url('PUT', key, 60, slash_safe=True)
        # 验证url中‘/’是否被转义
        seek = url.find(slash_safe_key)
        self.assertTrue(seek != -1)

        result = self.bucket.put_object_with_url(url, content)
        self.assertEqual(result.status, 200)

        result = self.bucket.head_object(key)
        self.assertEqual(result.content_length, 1024)

        result = self.bucket.delete_object(key)
        self.assertEqual(result.status, 204)

    def test_get_object_with_url_to_file_chunked(self):
        # object后缀为txt, length >= 1024, 指定Accept-Encoding接收，服务器将会以chunked模式传输数据
        key = 'test_get_object_with_url_to_file_chunked.txt'
        content = b'a' * 1024
        self.bucket.put_object(key, content)

        filename = key + '-local.txt'
        url = self.bucket.sign_url('GET', key, 240)
        result = self.bucket.get_object_with_url_to_file(url, filename, headers={'Accept-Encoding': 'gzip'})

        self.assertEqual(result.headers['Transfer-Encoding'], 'chunked')
        self.assertFileContent(filename, content)

        os.remove(filename)
        self.bucket.delete_object(key)

    def test_modified_since(self):
        key = self.random_key()
        content = random_bytes(16)

        self.bucket.put_object(key, content)
        self.assertRaises(oss2.exceptions.NotModified,
                          self.bucket.get_object,
                          key,
                          headers={
                              'if-modified-since': oss2.utils.http_date(int(time.time()) + 60),
                          },
                          byte_range=(0, 7))

    def test_copy_object(self):
        source_key = self.random_key()
        target_key = self.random_key()
        content = random_bytes(36)

        self.bucket.put_object(source_key, content)
        self.bucket.copy_object(self.bucket.bucket_name, source_key, target_key)

        result = self.bucket.get_object(target_key)
        self.assertEqual(content, result.read())

    def test_copy_object_source_with_escape(self):
        source_key = '阿里云/加油/:?;@&=+$, /<>{}[]|/'
        target_key = self.random_key()
        content = random_bytes(36)

        self.bucket.put_object(source_key, content)
        self.bucket.copy_object(self.bucket.bucket_name, source_key, target_key)

        result = self.bucket.get_object(target_key)
        self.assertEqual(content, result.read())

    def test_update_object_meta(self):
        key = self.random_key('.txt')
        content = random_bytes(36)

        self.bucket.put_object(key, content)

        # 更改Content-Type，增加用户自定义元数据
        self.bucket.update_object_meta(key, {'Content-Type': 'whatever',
                                                     'x-oss-meta-category': 'novel'})

        result = self.bucket.head_object(key)
        self.assertEqual(result.headers['content-type'], 'whatever')
        self.assertEqual(result.headers['x-oss-meta-category'], 'novel')

    def test_object_acl(self):
        key = self.random_key()
        content = random_bytes(32)

        self.bucket.put_object(key, content)
        self.assertEqual(self.bucket.get_object_acl(key).acl, oss2.OBJECT_ACL_DEFAULT)

        for permission in (oss2.OBJECT_ACL_PRIVATE, oss2.OBJECT_ACL_PUBLIC_READ, oss2.OBJECT_ACL_PUBLIC_READ_WRITE,
                           oss2.OBJECT_ACL_DEFAULT):
            self.bucket.put_object_acl(key, permission)
            self.assertEqual(self.bucket.get_object_acl(key).acl, permission)

        self.bucket.delete_object(key)

    def test_object_exists(self):
        key = self.random_key()
        
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-object-exists"
        bucket = oss2.Bucket(auth, OSS_ENDPOINT, bucket_name)
        self.assertRaises(NoSuchBucket, bucket.object_exists, key)

        self.assertTrue(not self.bucket.object_exists(key))

        self.bucket.put_object(key, "hello world")
        self.assertTrue(self.bucket.object_exists(key))

    def test_user_meta(self):
        key = self.random_key()

        self.bucket.put_object(key, 'hello', headers={'x-oss-meta-key1': 'value1',
                                                      'X-Oss-Meta-Key2': 'value2'})

        headers = self.bucket.get_object(key).headers
        self.assertEqual(headers['x-oss-meta-key1'], 'value1')
        self.assertEqual(headers['x-oss-meta-key2'], 'value2')
        
    def test_get_object_meta(self):
        key = self.random_key()
        content = 'hello'
        
        # bucket no exist
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-get-object-meta"
        bucket = oss2.Bucket(auth, OSS_ENDPOINT, bucket_name)
        
        self.assertRaises(NoSuchBucket, bucket.get_object_meta, key)
        
        # object no exist
        self.assertRaises(NoSuchKey, self.bucket.get_object_meta, key)

        self.bucket.put_object(key, content)
        
        # get meta normal
        result = self.bucket.get_object_meta(key)

        self.assertTrue(result.last_modified > 1472128796)
        self.assertEqual(result.content_length, len(content))
        self.assertEqual(result.etag, '5D41402ABC4B2A76B9719D911017C592')

    def test_progress(self):
        stats = {'previous': -1}

        def progress_callback(bytes_consumed, total_bytes):
            self.assertTrue(bytes_consumed <= total_bytes)
            self.assertTrue(bytes_consumed > stats['previous'])

            stats['previous'] = bytes_consumed

        key = self.random_key()
        content = random_bytes(2 * 1024 * 1024)

        # 上传内存中的内容
        stats = {'previous': -1}
        self.bucket.put_object(key, content, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        # 使用签名url上传
        stats = {'previous': -1}
        url = self.bucket.sign_url('PUT', key, 240)
        result = self.bucket.put_object_with_url(url, content, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        # 追加内容
        stats = {'previous': -1}
        self.bucket.append_object(self.random_key(), 0, content, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        # 下载到文件
        stats = {'previous': -1}
        filename = random_string(12) + '.txt'
        self.bucket.get_object_to_file(key, filename, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        # 上传本地文件
        stats = {'previous': -1}
        self.bucket.put_object_from_file(key, filename, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        # 下载到本地，采用iterator语法
        stats = {'previous': -1}
        result = self.bucket.get_object(key, progress_callback=progress_callback)
        content_got = b''
        for chunk in result:
            content_got += chunk
        self.assertEqual(stats['previous'], len(content))
        self.assertEqual(content, content_got)

        os.remove(filename)

    def test_exceptions(self):
        key = self.random_key()
        content = random_bytes(16)

        self.assertRaises(NotFound, self.bucket.get_object, key)
        self.assertRaises(NoSuchKey, self.bucket.get_object, key)

        self.bucket.put_object(key, content)

        self.assertRaises(Conflict, self.bucket.append_object, key, len(content), b'more content')
        self.assertRaises(ObjectNotAppendable, self.bucket.append_object, key, len(content), b'more content')

    def test_gzip_get(self):
        #"""OSS supports HTTP Compression, see https://en.wikipedia.org/wiki/HTTP_compression for details.
        #"""
        key = self.random_key('.txt')       # ensure our content-type is text/plain, which could be compressed
        content = random_bytes(1024 * 1024) # ensure our content-length is larger than 1024 to trigger compression

        self.bucket.put_object(key, content)

        result = self.bucket.get_object(key, headers={'Accept-Encoding': 'gzip'})
        content_read = result.read()
        self.assertEqual(len(content_read), len(content))
        self.assertEqual(content_read, content)
        self.assertTrue(result.content_length is None)
        self.assertEqual(result.headers['Content-Encoding'], 'gzip')

        # test progress
        stats = {'previous': -1}

        def progress_callback(bytes_consumed, total_bytes):
            self.assertTrue(total_bytes is None)
            self.assertTrue(bytes_consumed > stats['previous'])
            stats['previous'] = bytes_consumed

        content_got = b''
        result = self.bucket.get_object(key, headers={'Accept-Encoding': 'gzip'}, progress_callback=progress_callback)
        content_got = result.read()

        self.assertEqual(len(content), len(content_got))
        self.assertEqual(content, content_got)

    def test_invalid_object_name(self):
        key = '\invalid-object-name'
        content = random_bytes(16)

        self.assertRaises(oss2.exceptions.InvalidObjectName, self.bucket.put_object, key, content)

    def test_disable_crc(self): 
        key = self.random_key('.txt')
        content = random_bytes(1024 * 100)
        
        bucket = oss2.Bucket(oss2.Auth(OSS_ID, OSS_SECRET), OSS_ENDPOINT, OSS_BUCKET, enable_crc=False)
        
        # put
        put_result = bucket.put_object(key, content)
        self.assertFalse(hasattr(put_result, 'get_crc'))
        self.assertTrue(put_result.crc is not None)
        
        # get 
        get_result = bucket.get_object(key)
        self.assertEqual(get_result.read(), content)
        self.assertTrue(get_result.client_crc is None)
        self.assertTrue(get_result.server_crc)
        
        bucket.delete_object(key)
        
        # append
        append_result = bucket.append_object(key, 0, content)
        self.assertFalse(hasattr(append_result, 'get_crc'))
        self.assertTrue(append_result.crc is not None)
        
        append_result = bucket.append_object(key, len(content), content)
        self.assertFalse(hasattr(append_result, 'get_crc'))
        self.assertTrue(append_result.crc is not None)
        
        bucket.delete_object(key)
        
        # multipart
        upload_id = bucket.init_multipart_upload(key).upload_id

        parts = []
        result = bucket.upload_part(key, upload_id, 1, content)
        parts.append(oss2.models.PartInfo(1, result.etag))
        result = bucket.upload_part(key, upload_id, 2, content)
        parts.append(oss2.models.PartInfo(2, result.etag))

        bucket.complete_multipart_upload(key, upload_id, parts)

    def test_invalid_crc(self):
        key = self.random_key()
        content = random_bytes(512)

        try:
            self.bucket.append_object(key, 0, content, init_crc=1)
        except oss2.exceptions.InconsistentError as e:
            self.assertEqual(e.status, -3)
            self.assertTrue(e.body.startswith('InconsistentError'))
        else:
            self.assertTrue(False)

    def test_put_symlink(self):
        key  = self.random_key()
        symlink = self.random_key()
        content = 'hello'
        
        self.bucket.put_object(key, content)
        
        # put symlink normal
        self.bucket.put_symlink(key, symlink)
        
        head_result = self.bucket.head_object(symlink)
        self.assertEqual(head_result.content_length, len(content))
        self.assertEqual(head_result.etag, '5D41402ABC4B2A76B9719D911017C592')

        self.bucket.put_object(key, content)
        
        # put symlink with meta
        self.bucket.put_symlink(key, symlink, headers={'x-oss-meta-key1': 'value1',
                                                              'X-Oss-Meta-Key2': 'value2'})
        head_result = self.bucket.head_object(symlink)
        self.assertEqual(head_result.content_length, len(content))
        self.assertEqual(head_result.etag, '5D41402ABC4B2A76B9719D911017C592')
        self.assertEqual(head_result.headers['x-oss-meta-key1'], 'value1')
        self.assertEqual(head_result.headers['x-oss-meta-key2'], 'value2')

    def test_get_symlink(self):
        key = self.random_key()
        symlink = self.random_key()
        content = 'hello'
        
        # bucket no exist
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-get-symlink"
        bucket = oss2.Bucket(auth, OSS_ENDPOINT, bucket_name)
        
        self.assertRaises(NoSuchBucket, bucket.get_symlink, symlink)
        
        # object no exist
        self.assertRaises(NoSuchKey, self.bucket.get_symlink, symlink)
        
        self.bucket.put_object(key, content)
        self.bucket.put_symlink(key, symlink)
        
        # get symlink normal
        result = self.bucket.get_symlink(symlink)
        self.assertEqual(result.target_key, key)

    def test_process_object(self):
        key = self.random_key(".jpg")
        result = self.bucket.put_object_from_file(key, "tests/example.jpg")
        self.assertEqual(result.status, 200)

        dest_key = self.random_key(".jpg")
        process = "image/resize,w_100|sys/saveas,o_{0},b_{1}".format(
            oss2.compat.to_string(base64.urlsafe_b64encode(oss2.compat.to_bytes(dest_key))),
            oss2.compat.to_string(base64.urlsafe_b64encode(oss2.compat.to_bytes(self.bucket.bucket_name))))
        result = self.bucket.process_object(key, process)
        self.assertEqual(result.status, 200)
        self.assertEqual(result.bucket, self.bucket.bucket_name)
        self.assertEqual(result.object, dest_key)
        result = self.bucket.object_exists(dest_key)
        self.assertEqual(result, True)


        # If bucket-name not specified, it is saved to the current bucket by default.
        dest_key = self.random_key(".jpg")
        process = "image/resize,w_100|sys/saveas,o_{0}".format(
            oss2.compat.to_string(base64.urlsafe_b64encode(oss2.compat.to_bytes(dest_key))))
        result = self.bucket.process_object(key, process)
        self.assertEqual(result.status, 200)
        self.assertEqual(result.bucket, "")
        self.assertEqual(result.object, dest_key)
        result = self.bucket.object_exists(dest_key)
        self.assertEqual(result, True)
    
    def test_object_tagging_client_error(self):

        rule = TaggingRule()
        self.assertRaises(oss2.exceptions.ClientError, rule.add, 129*'a', 'test')
        self.assertRaises(oss2.exceptions.ClientError, rule.add, 'test', 257*'a')
        self.assertRaises(oss2.exceptions.ClientError, rule.add, None, 'test')
        self.assertRaises(oss2.exceptions.ClientError, rule.add, '', 'test')
        self.assertRaises(KeyError, rule.delete, 'not_exist')

    def test_object_tagging_wrong_key(self):
       
        tagging = Tagging()
        tagging.tag_set.tagging_rule[129*'a'] = 'test'
        
        key = self.random_key('.dat')
        result = self.bucket.put_object(key, "test")

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

        tagging.tag_set.delete(129*'a')

        self.assertTrue( 129*'a' not in tagging.tag_set.tagging_rule )

        tagging.tag_set.tagging_rule['%@abc'] = 'abc'

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

        tagging.tag_set.delete('%@abc')

        self.assertTrue( '%@abc' not in tagging.tag_set.tagging_rule )

        tagging.tag_set.tagging_rule[''] = 'abc'

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass


    def test_object_tagging_wrong_value(self):
       
        tagging = Tagging()

        tagging.tag_set.tagging_rule['test'] = 257*'a'

        key = self.random_key('.dat')

        result = self.bucket.put_object(key, "test")

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

        tagging.tag_set.tagging_rule['test']= '%abc'

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

        tagging.tag_set.tagging_rule['test']= ''

        try:
            result = self.bucket.put_object_tagging(key, tagging)
        except oss2.exceptions.OssError:
            self.assertFalse(True, 'should get exception')

    def test_object_tagging_wrong_rule_num(self):
        
        key = self.random_key('.dat')
        result = self.bucket.put_object(key, "test")

        tagging = Tagging(None)
        for i in range(0,12):
            key='test_'+str(i)
            value='test_'+str(i)
            tagging.tag_set.add(key, value)

        try:
            result = self.bucket.put_object_tagging(key, tagging)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

    def test_object_tagging(self):

        key = self.random_key('.dat')
        result = self.bucket.put_object(key, "test")

        try:
            result=self.bucket.get_object_tagging(key)
            self.assertEqual(0, result.tag_set.len())
        except oss2.exceptions.OssError:
            self.assertFalse(True, "should get exception")

        rule = TaggingRule()
        key1=128*'a'
        value1=256*'a'
        rule.add(key1, value1)

        key2='+-.:'
        value2='_/'
        rule.add(key2, value2)

        tagging = Tagging(rule) 
        result = self.bucket.put_object_tagging(key, tagging)
        self.assertTrue(200, result.status)

        result = self.bucket.get_object_tagging(key)

        self.assertEqual(2, result.tag_set.len())
        self.assertEqual(256*'a', result.tag_set.tagging_rule[128*'a'])
        self.assertEqual('_/', result.tag_set.tagging_rule['+-.:'])

    def test_put_object_with_tagging(self):
    
        key = self.random_key('.dat')
        headers = dict()
        headers[OSS_OBJECT_TAGGING] = "k1=v1&k2=v2&k3=v3"

        result = self.bucket.put_object(key, 'test', headers=headers)
        self.assertEqual(200, result.status)
        
        result = self.bucket.get_object_tagging(key)
        self.assertEqual(3, result.tag_set.len())
        self.assertEqual('v1', result.tag_set.tagging_rule['k1'])
        self.assertEqual('v2', result.tag_set.tagging_rule['k2'])
        self.assertEqual('v3', result.tag_set.tagging_rule['k3'])

        result = self.bucket.delete_object_tagging(key)

        self.assertEqual(204, result.status)

        result = self.bucket.get_object_tagging(key)
        self.assertEqual(0, result.tag_set.len())

    def test_copy_object_with_tagging(self):
    
        #key = self.random_key('.dat')
        key = 'aaaaaaaaaaaaaa' 
        headers = dict()
        headers[OSS_OBJECT_TAGGING] = "k1=v1&k2=v2&k3=v3"

        result = self.bucket.put_object(key, 'test', headers=headers)
        self.assertEqual(200, result.status)

        result = self.bucket.get_object_tagging(key)
        self.assertEqual(3, result.tag_set.len())
        self.assertEqual('v1', result.tag_set.tagging_rule['k1'])
        self.assertEqual('v2', result.tag_set.tagging_rule['k2'])
        self.assertEqual('v3', result.tag_set.tagging_rule['k3'])

        headers=dict()
        headers[OSS_OBJECT_TAGGING_COPY_DIRECTIVE] = 'COPY'
        result = self.bucket.copy_object(self.bucket.bucket_name, key, key+'_test', headers=headers)

        result = self.bucket.get_object_tagging(key+'_test')
        self.assertEqual(3, result.tag_set.len())
        self.assertEqual('v1', result.tag_set.tagging_rule['k1'])
        self.assertEqual('v2', result.tag_set.tagging_rule['k2'])
        self.assertEqual('v3', result.tag_set.tagging_rule['k3'])

        tag_key1 = u' +/ '
        tag_value1 = u'中文'
        tag_str = urlquote(tag_key1.encode('UTF-8')) + '=' + urlquote(tag_value1.encode('UTF-8'))

        tag_key2 = u'中文'
        tag_value2 = u'test++/'
        tag_str += '&' + urlquote(tag_key2.encode('UTF-8')) + '=' + urlquote(tag_value2.encode('UTF-8'))

        headers[OSS_OBJECT_TAGGING] = tag_str
        headers[OSS_OBJECT_TAGGING_COPY_DIRECTIVE] = 'REPLACE'
        result = self.bucket.copy_object(self.bucket.bucket_name, key, key+'_test', headers=headers)

        result = self.bucket.get_object_tagging(key+'_test')
        self.assertEqual(2, result.tag_set.len())
        self.assertEqual('中文', result.tag_set.tagging_rule[' +/ '])
        self.assertEqual('test++/', result.tag_set.tagging_rule['中文'])

    def test_append_object_with_tagging(self):
        key = self.random_key()
        content1 = random_bytes(512)
        content2 = random_bytes(128)

        result = self.bucket.append_object(key, 0, content1, init_crc=0)
        self.assertEqual(result.next_position, len(content1))
        self.assertTrue(result.crc is not None)

        try:
            self.bucket.append_object(key, 0, content2)
        except PositionNotEqualToLength as e:
            self.assertEqual(e.next_position, len(content1))
        else:
            self.assertTrue(False)
        
        result = self.bucket.append_object(key, len(content1), content2, init_crc=result.crc)
        self.assertEqual(result.next_position, len(content1) + len(content2))
        self.assertTrue(result.crc is not None)

        self.bucket.delete_object(key)

        rule = TaggingRule()
        self.assertEqual('', rule.to_query_string())

        rule.add('key1', 'value1')
        self.assertEqual(rule.to_query_string(), 'key1=value1')

        rule.add(128*'a', 256*'b')
        rule.add('+-/', ':+:')
        self.assertEqual('key1=value1' in rule.to_query_string(), True)
        self.assertEqual((128*'a' + '=' + 256*'b') in rule.to_query_string(), True)
        self.assertEqual('%2B-/=%3A%2B%3A' in rule.to_query_string(), True)

        headers = dict()
        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()

        result = self.bucket.append_object(key, 0, content1, init_crc=0, headers=headers)
        self.assertEqual(result.next_position, len(content1))
        self.assertTrue(result.crc is not None)

        result = self.bucket.append_object(key, len(content1), content2, init_crc=result.crc)
        self.assertEqual(result.next_position, len(content1) + len(content2))
        self.assertTrue(result.crc is not None)

        result = self.bucket.get_object_tagging(key)
        self.assertEqual(3, result.tag_set.len())

        tagging_rule = result.tag_set.tagging_rule
        self.assertEqual('value1', tagging_rule['key1'])
        self.assertEqual(256*'b', tagging_rule[128*'a'])
        self.assertEqual(':+:', tagging_rule['+-/'])

    def test_append_object_with_tagging_wrong_num(self):
        key = self.random_key()
        content1 = random_bytes(512)
        content2 = random_bytes(128)

        result = self.bucket.append_object(key, 0, content1, init_crc=0)
        self.assertEqual(result.next_position, len(content1))
        self.assertTrue(result.crc is not None)

        try:
            self.bucket.append_object(key, 0, content2)
        except PositionNotEqualToLength as e:
            self.assertEqual(e.next_position, len(content1))
        else:
            self.assertTrue(False)
        
        result = self.bucket.append_object(key, len(content1), content2, init_crc=result.crc)
        self.assertEqual(result.next_position, len(content1) + len(content2))
        self.assertTrue(result.crc is not None)

        self.bucket.delete_object(key)
        
        # append object with wrong tagging kv num, but not in 
        # first call, it will be ignored
        rule = TaggingRule()
        self.assertEqual('', rule.to_query_string())

        for i in range(0, 15):
            tag_key = 'key' + str(i)
            tag_value = 'value' + str(i)
            rule.add(tag_key, tag_value)

        headers = dict()
        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()


        result = self.bucket.append_object(key, 0, content1, init_crc=0)
        self.assertEqual(result.next_position, len(content1))
        self.assertTrue(result.crc is not None)

        result = self.bucket.append_object(key, len(content1), content2, init_crc=result.crc, headers=headers)

        result_tagging = self.bucket.get_object_tagging(key)
        self.assertEqual(0, result_tagging.tag_set.len())
        
        rule.delete('key1')
        rule.delete('key2')
        rule.delete('key3')
        rule.delete('key4')
        rule.delete('key5')
        rule.delete('key6')

        self.assertEqual(9, rule.len())

        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()

        try:
            result = self.bucket.append_object(key, len(content1)+len(content2), 
                    content2, init_crc=result.crc, headers=headers)
        except oss2.exceptions.OssError:
            self.assertFalse(True, 'should not get exception')

        result = self.bucket.get_object_tagging(key)
        self.assertEqual(0, result.tag_set.len())

        self.bucket.delete_object(key)

        wait_meta_sync()

        # append object with wrong tagging kv num in first call,
        # it will be fail
        rule = TaggingRule()
        self.assertEqual('', rule.to_query_string())

        for i in range(0, 15):
            tag_key = 'key' + str(i)
            tag_value = 'value' + str(i)
            rule.add(tag_key, tag_value)

        headers = dict()
        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()

        try:
            self.bucket.append_object(key, 0, content1, init_crc=0, headers=headers)
            self.assertFalse(True, 'should get exception')
        except oss2.exceptions.OssError:
            pass

    def test_put_symlink_with_tagging(self):
        key  = self.random_key()
        symlink = self.random_key()
        content = 'hello'
        
        self.bucket.put_object(key, content)
        
        rule = TaggingRule()
        self.assertEqual('', rule.to_query_string())

        rule.add('key1', 'value1')
        self.assertTrue(rule.to_query_string() != '')

        rule.add(128*'a', 256*'b')
        rule.add('+-/', ':+:')

        headers = dict()
        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()

        # put symlink normal
        self.bucket.put_symlink(key, symlink, headers=headers)

        result = self.bucket.get_object_tagging(symlink)
        self.assertEqual(3, result.tag_set.len())

        tagging_rule = result.tag_set.tagging_rule
        self.assertEqual('value1', tagging_rule['key1'])
        self.assertEqual(256*'b', tagging_rule[128*'a'])
        self.assertEqual(':+:', tagging_rule['+-/'])

        result = self.bucket.delete_object(symlink)
        self.assertEqual(204, int(result.status))

    def test_put_symlink_with_tagging_with_wrong_num(self):
        key  = self.random_key()
        symlink = self.random_key()
        content = 'hello'
        self.bucket.put_object(key, content)
        
        rule = TaggingRule()
        self.assertEqual('', rule.to_query_string())

        for i in range(0, 15):
            tag_key = 'key' + str(i)
            tag_value = 'value' + str(i)
            rule.add(tag_key, tag_value)

        headers = dict()
        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()
        
        try:
            self.bucket.put_symlink(key, symlink, headers=headers)
            self.assertFalse(True, 'should get exception')
        except:
            pass
       
        rule.delete('key1')
        rule.delete('key2')
        rule.delete('key3')
        rule.delete('key4')
        rule.delete('key5')
        rule.delete('key6')

        headers[OSS_OBJECT_TAGGING] = rule.to_query_string()

        try:
            result = self.bucket.put_symlink(key, symlink, headers=headers)
        except:
            self.assertFalse(True, 'should not get exception')

        head_result = self.bucket.head_object(symlink)
        self.assertEqual(head_result.content_length, len(content))
        self.assertEqual(head_result.etag, '5D41402ABC4B2A76B9719D911017C592')

        # put symlink with meta
        self.bucket.put_symlink(key, symlink, headers={'x-oss-meta-key1': 'value1',
                'x-oss-meta-KEY2': 'value2'})

        head_result = self.bucket.head_object(symlink)
        self.assertEqual(head_result.content_length, len(content))
        self.assertEqual(head_result.etag, '5D41402ABC4B2A76B9719D911017C592')
        self.assertEqual(head_result.headers['x-oss-meta-key1'], 'value1')
        self.assertEqual(head_result.headers['x-oss-meta-key2'], 'value2')

    def test_put_object_with_unicode_header(self):
        key = self.random_key()
        value = '测试'
        byte = to_bytes(value)
        headers={'x-oss-meta-unicode': byte}
        self.bucket.put_object(key, 'a novel', headers=headers)
        result = self.bucket.head_object(key)
        self.assertEqual(result.status, 200)
        newstr = result.headers['x-oss-meta-unicode']
        if is_py2:
            b_str = newstr
        else:
            b_str = newstr.encode('iso-8859-1')
        self.assertEqual(to_string(b_str), value)

    def test_with_get_object_result(self):
        key = 'test-with-get-object-result'
        content = b'1' * 1024
        self.bucket.put_object(key, content)
        result = self.bucket.get_object(key)
        with result as f:
            data = f.read()
            self.assertEqual(content, data)

        result = self.bucket.get_object(key)
        with result as f:
            pass
        data = result.read()
        self.assertEqual(0, len(data))

    
class TestSign(TestObject):
    """
        这个类主要是用来增加测试覆盖率，当环境变量为oss2.AUTH_VERSION_2，则重新设置为oss2.AUTH_VERSION_1再运行TestObject，反之亦然
    """
    def __init__(self, *args, **kwargs):
        super(TestSign, self).__init__(*args, **kwargs)

    def setUp(self):
        if os.getenv('OSS_TEST_AUTH_VERSION') == oss2.AUTH_VERSION_2:
            os.environ['OSS_TEST_AUTH_VERSION'] = oss2.AUTH_VERSION_1
        else:
            os.environ['OSS_TEST_AUTH_VERSION'] = oss2.AUTH_VERSION_2
        super(TestSign, self).setUp()

    def tearDown(self):
        if os.getenv('OSS_TEST_AUTH_VERSION') == oss2.AUTH_VERSION_2:
            os.environ['OSS_TEST_AUTH_VERSION'] = oss2.AUTH_VERSION_1
        else:
            os.environ['OSS_TEST_AUTH_VERSION'] = oss2.AUTH_VERSION_2
        super(TestSign, self).tearDown()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import unittest
import oss2
from oss2.utils import calc_obj_crc_from_parts

from .common import *
from oss2.headers import OSS_OBJECT_TAGGING, OSS_OBJECT_TAGGING_COPY_DIRECTIVE
from oss2.compat import urlunquote, urlquote

from oss2.compat import is_py2, is_py33


class TestMultipart(OssTestCase):
    def do_multipart_internal(self, do_md5):
        key = self.random_key()
        content = random_bytes(128 * 1024)

        parts = []
        upload_id = self.bucket.init_multipart_upload(key).upload_id

        if do_md5:
            headers = {'Content-Md5': oss2.utils.content_md5(content)}
        else:
            headers = None

        result = self.bucket.upload_part(key, upload_id, 1, content, headers=headers)
        parts.append(oss2.models.PartInfo(1, result.etag, size=len(content), part_crc=result.crc))
        self.assertTrue(result.crc is not None)

        complete_result = self.bucket.complete_multipart_upload(key, upload_id, parts)

        object_crc = calc_obj_crc_from_parts(parts)
        self.assertTrue(complete_result.crc is not None)
        self.assertEqual(object_crc, result.crc)

        result = self.bucket.get_object(key)
        self.assertEqual(content, result.read())

    def test_multipart(self):
        self.do_multipart_internal(False)

    def test_upload_part_content_md5_good(self):
        self.do_multipart_internal(True)

    def test_upload_part_content_md5_bad(self):
        key = self.random_key()
        content = random_bytes(128 * 1024)

        parts = []
        upload_id = self.bucket.init_multipart_upload(key).upload_id

        # construct a bad Content-Md5 by using 'content + content's Content-Md5
        headers = {'Content-Md5': oss2.utils.content_md5(content + content)}

        self.assertRaises(oss2.exceptions.InvalidDigest, self.bucket.upload_part, key, upload_id, 1, content, headers=headers)

    def test_progress(self):
        stats = {'previous': -1}

        def progress_callback(bytes_consumed, total_bytes):
            self.assertTrue(bytes_consumed <= total_bytes)
            self.assertTrue(bytes_consumed > stats['previous'])

            stats['previous'] = bytes_consumed

        key = self.random_key()
        content = random_bytes(128 * 1024)

        upload_id = self.bucket.init_multipart_upload(key).upload_id
        self.bucket.upload_part(key, upload_id, 1, content, progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        self.bucket.abort_multipart_upload(key, upload_id)

    def test_upload_part_copy(self):
        src_object = self.random_key()
        dst_object = self.random_key()

        content = random_bytes(200 * 1024)

        # 上传源文件
        self.bucket.put_object(src_object, content)

        # part copy到目标文件
        parts = []
        upload_id = self.bucket.init_multipart_upload(dst_object).upload_id

        result = self.bucket.upload_part_copy(self.bucket.bucket_name, src_object,
                                              (0, 100 * 1024 - 1), dst_object, upload_id, 1)
        parts.append(oss2.models.PartInfo(1, result.etag))

        result = self.bucket.upload_part_copy(self.bucket.bucket_name, src_object,
                                              (100*1024, None), dst_object, upload_id, 2)
        parts.append(oss2.models.PartInfo(2, result.etag))

        self.bucket.complete_multipart_upload(dst_object, upload_id, parts)

        # 验证
        content_got = self.bucket.get_object(dst_object).read()
        self.assertEqual(len(content_got), len(content))
        self.assertEqual(content_got, content)

    def test_init_multipart_with_object_tagging_exceptions(self):

        key = self.random_key()

        headers=dict()
        # wrong key
        tag_str='=a&b=a'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
            self.assertFalse(True, 'should get a exception')
        except oss2.exceptions.OssError:
            pass

        # wrong key
        long_str=129*'a'
        tag_str=long_str+'=b&b=a'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
            self.assertFalse(True, 'should get a exception')
        except oss2.exceptions.OssError:
            pass


        # wrong value
        tag_str='a=&b=c'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
        except oss2.exceptions.OssError:
            self.assertFalse(True, 'should get a exception')


        # wrong value
        long_str=257*'a'
        tag_str = 'a='+long_str+'&b=a'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
            self.assertFalse(True, 'should get a exception')
        except oss2.exceptions.OssError:
            pass


        # dup kv
        tag_str='a=b&a=b&a=b'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
            self.assertFalse(True, 'should get a exception')
        except oss2.exceptions.OssError:
            pass


        # max+1 kv pairs
        tag_str='a1=b1&a2=b2&a3=b4&a4=b4&a5=b5&a6=b6&a7=b7&a8=b8&a9=b9&a10=b10&a11=b11&a12=b12'
        headers[OSS_OBJECT_TAGGING] = tag_str
        try:
            resp = self.bucket.init_multipart_upload(key, headers=headers)
            self.assertFalse(True, 'should get a exception')
        except oss2.exceptions.OssError:
            pass

    def test_multipart_with_object_tagging(self):

        key = self.random_key()
        content = random_bytes(128 * 1024)

        tag_str=''

        tag_key1=urlquote('+:/')
        tag_value1=urlquote('.-')
        tag_str = tag_key1+'='+tag_value1

        tag_ke2=urlquote(' + ')
        tag_value2=urlquote(u'中文'.encode('UTF-8'))
        tag_str += '&'+tag_ke2+'='+tag_value2

        headers=dict()
        headers[OSS_OBJECT_TAGGING] = tag_str

        parts = []
        upload_id = self.bucket.init_multipart_upload(key, headers=headers).upload_id

        headers = {'Content-Md5': oss2.utils.content_md5(content)}

        result = self.bucket.upload_part(key, upload_id, 1, content, headers=headers)
        parts.append(oss2.models.PartInfo(1, result.etag, size=len(content), part_crc=result.crc))
        self.assertTrue(result.crc is not None)

        complete_result = self.bucket.complete_multipart_upload(key, upload_id, parts)

        object_crc = calc_obj_crc_from_parts(parts)
        self.assertTrue(complete_result.crc is not None)
        self.assertEqual(object_crc, result.crc)

        result = self.bucket.get_object(key)
        self.assertEqual(content, result.read())

        result = self.bucket.get_object_tagging(key)

        self.assertEqual(2, result.tag_set.len())
        self.assertEqual('.-', result.tag_set.tagging_rule['+:/'])
        self.assertEqual('中文', result.tag_set.tagging_rule[' + '])

        result = self.bucket.delete_object_tagging(key)

    def test_multipart_sequential(self): 
        endpoint = "http://oss-cn-shanghai.aliyuncs.com"
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-multipart-sequential"
        bucket = oss2.Bucket(auth, endpoint, bucket_name)
        bucket.create_bucket()

        key = self.random_key()
        content = random_bytes(128 * 1024)

        parts = []
        upload_id = bucket.init_multipart_upload(key).upload_id
        result = bucket.upload_part(key, upload_id, 1, content)
        parts.append(oss2.models.PartInfo(1, result.etag, size=len(content), part_crc=result.crc))
        bucket.complete_multipart_upload(key, upload_id, parts)
        
        result = bucket.get_object(key)
        self.assertIsNone(result.resp.headers.get('Content-MD5'))

        parts = []
        params = {'sequential':''}
        upload_id = bucket.init_multipart_upload(key, params=params).upload_id
        result = bucket.upload_part(key, upload_id, 1, content)
        parts.append(oss2.models.PartInfo(1, result.etag, size=len(content), part_crc=result.crc))
        bucket.complete_multipart_upload(key, upload_id, parts)

        result = bucket.get_object(key)
        self.assertIsNotNone(result.resp.headers.get('Content-MD5'))

        bucket.delete_object(key)
        bucket.delete_bucket()


if __name__ == '__main__':
    unittest.main()



from .common import *

class TestBucketWorm(OssTestCase):
    def test_bucke_worm_normal(self):
        init_result = self.bucket.init_bucket_worm(1)
        worm_id = init_result.worm_id
        self.assertIsNotNone(init_result.request_id)

        get_result = self.bucket.get_bucket_worm()
        self.assertIsNotNone(get_result.request_id)
        self.assertEqual(worm_id, get_result.worm_id)
        self.assertEqual('InProgress', get_result.state)
        self.assertEqual(1, get_result.retention_period_days)
        self.assertIsNotNone(get_result.creation_date)

        complete_reuslt = self.bucket.complete_bucket_worm(worm_id)
        self.assertIsNotNone(complete_reuslt.request_id)
        get_result = self.bucket.get_bucket_worm()
        self.assertEqual(worm_id, get_result.worm_id)
        self.assertEqual('Locked', get_result.state)
        self.assertEqual(1, get_result.retention_period_days)
        self.assertIsNotNone(get_result.creation_date)

        extend_result = self.bucket.extend_bucket_worm(worm_id, 2)
        self.assertIsNotNone(extend_result.request_id)
        get_result = self.bucket.get_bucket_worm()
        self.assertEqual(worm_id, get_result.worm_id)
        self.assertEqual('Locked', get_result.state)
        self.assertEqual(2, get_result.retention_period_days)
        self.assertIsNotNone(get_result.creation_date)

    def test_abort_bucket_worm(self):
        self.bucket.init_bucket_worm(1)
        abort_result = self.bucket.abort_bucket_worm()
        self.assertIsNotNone(abort_result.request_id)

        init_result = self.bucket.init_bucket_worm(1)
        worm_id = init_result.worm_id

        self.bucket.complete_bucket_worm(worm_id)
        self.assertRaises(oss2.exceptions.WORMConfigurationLocked, self.bucket.abort_bucket_worm)


    def test_bucket_worm_illegal(self):
        self.assertRaises(oss2.exceptions.NoSuchWORMConfiguration, self.bucket.get_bucket_worm)

        init_result = self.bucket.init_bucket_worm(1)
        worm_id = init_result.worm_id

        self.assertRaises(oss2.exceptions.InvalidWORMConfiguration, self.bucket.extend_bucket_worm, worm_id, 2)


if __name__ == '__main__':
        unittest.main()



# -*- coding: utf-8 -*-

import unittest
import sys
import oss2
import json

from .common import *


class TestImage(OssTestCase):
    def __prepare(self):
        original_image = 'tests/example.jpg'
        new_image = self.random_key('.jpg')
        
        self.bucket.put_object_from_file(original_image, original_image)
        
        return original_image, new_image
    
    def __clear(self, original_image, new_image):
        self.bucket.delete_object(original_image)
        self.bucket.delete_object(new_image)

    def __test(self, original_image, new_image, image_style):
        original_image_content = self.bucket.get_object(original_image, process=image_style)
        self.bucket.put_object(new_image, original_image_content)
        
    def __test_to_file(self, original_image, new_image, image_style):
        self.bucket.get_object_to_file(original_image, new_image, process=image_style)
        self.bucket.put_object_from_file(new_image, new_image)
        oss2.utils.silently_remove(new_image)
        
    def __check(self, image_key, image_height, image_width, image_size, image_format):
        result = self.bucket.get_object(image_key, process='image/info')
        json_content = result.read()
        decoded_json = json.loads(oss2.to_unicode(json_content))
        
        self.assertEqual(int(decoded_json['ImageHeight']['value']), image_height)
        self.assertEqual(int(decoded_json['ImageWidth']['value']), image_width)
        self.assertEqual(int(decoded_json['FileSize']['value']), image_size)
        self.assertEqual(decoded_json['Format']['value'], image_format)
    
    def test_resize(self):
        style = "image/resize,m_fixed,w_100,h_100"  # 缩放
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 100, 100, 3267, 'jpg')
        self.__clear(original_image, new_image)
            
    def test_crop(self):
        style = "image/crop,w_100,h_100,x_100,y_100,r_1"  # 裁剪
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 100, 100, 1969, 'jpg')
        self.__clear(original_image, new_image)
        
    def test_rotate(self):
        style = "image/rotate,90"  # 旋转
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 400, 267, 20998, 'jpg')
        self.__clear(original_image, new_image)
        
    def test_sharpen(self):
        style = "image/sharpen,100"  # 锐化
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 267, 400, 23015, 'jpg')
        self.__clear(original_image, new_image)
        
    def test_watermark(self):
        style = "image/watermark,text_SGVsbG8g5Zu-54mH5pyN5YqhIQ"  # 文字水印
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 267, 400, 26378, 'jpg')
        self.__clear(original_image, new_image)
        
    def test_format(self):
        style = "image/format,png"  # 图像格式转换
        
        original_image, new_image = self.__prepare()
        self.__test(original_image, new_image, style)
        self.__check(new_image, 267, 400, 160733, 'png')
        self.__clear(original_image, new_image)
        
    def test_resize_to_file(self):
        style = "image/resize,m_fixed,w_100,h_100"  # 缩放
        
        original_image, new_image = self.__prepare()
        self.__test_to_file(original_image, new_image, style)
        self.__check(new_image, 100, 100, 3267, 'jpg')
        self.__clear(original_image, new_image)
         

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os
import oss2

# 以下代码展示了set_bucket_storage_capacity的的用法。

# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
access_key_id = os.getenv('OSS_TEST_ACCESS_KEY_ID', '<你的AccessKeyId>')
access_key_secret = os.getenv('OSS_TEST_ACCESS_KEY_SECRET', '<你的AccessKeySecret>')
bucket_name = os.getenv('OSS_TEST_BUCKET', '<你的Bucket>')
endpoint = os.getenv('OSS_TEST_ENDPOINT', '<你的访问域名>')

# 确认上面的参数都填写正确了
for param in (access_key_id, access_key_secret, bucket_name, endpoint):
    assert '<' not in param, '请设置参数：' + param

# 创建Bucket对象，所有Object相关的接口都可以通过Bucket对象来进行
bucket = oss2.Bucket(oss2.Auth(access_key_id, access_key_secret), endpoint, bucket_name)

# 设置bucket容量为100GB
user_qos = oss2.models.BucketUserQos(100)
resp = bucket.set_bucket_storage_capacity(user_qos)

# 获取容量信息
result = bucket.get_bucket_storage_capacity()
print("bucket_qos_capacity:" + result.storage_capacity)


# -*- coding: utf-8 -*-

import unittest
import oss2
import os
import sys
import time
from oss2 import models
from oss2 import utils

from .common import *

from mock import patch


class TestUpload(OssTestCase):
    def test_upload_small(self):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        key = random_string(16)
        content = random_bytes(100)

        pathname = self._prepare_temp_file(content)

        result = oss2.resumable_upload(bucket, key, pathname)
        self.assertTrue(result is not None)
        self.assertTrue(result.etag is not None)
        self.assertTrue(result.request_id is not None)

        result = bucket.get_object(key)
        self.assertEqual(content, result.read())
        self.assertEqual(result.headers['x-oss-object-type'], 'Normal')

        bucket.delete_object(key)

    def test_upload_large(self):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        key = random_string(16)
        content = random_bytes(500 * 1024)

        pathname = self._prepare_temp_file(content)

        result = oss2.resumable_upload(bucket, key, pathname, multipart_threshold=200 * 1024, part_size=None)
        self.assertTrue(result is not None)
        self.assertTrue(result.etag is not None)
        self.assertTrue(result.request_id is not None)

        result = bucket.get_object(key)
        self.assertEqual(content, result.read())
        self.assertEqual(result.headers['x-oss-object-type'], 'Multipart')

        bucket.delete_object(key)

    def test_concurrency(self):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        key = random_string(16)
        content = random_bytes(64 * 100 * 1024)

        pathname = self._prepare_temp_file(content)

        oss2.resumable_upload(bucket, key, pathname, multipart_threshold=200 * 1024, part_size=100 * 1024,
                              num_threads=8)
        result = bucket.get_object(key)
        self.assertEqual(content, result.read())
        self.assertEqual(result.headers['x-oss-object-type'], 'Multipart')

    def test_progress(self):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        stats = {'previous': -1, 'ncalled': 0}

        def progress_callback(bytes_consumed, total_bytes):
            self.assertTrue(bytes_consumed <= total_bytes)
            self.assertTrue(bytes_consumed > stats['previous'])

            stats['previous'] = bytes_consumed
            stats['ncalled'] += 1

        key = random_string(16)
        content = random_bytes(5 * 100 * 1024 + 100)

        pathname = self._prepare_temp_file(content)

        part_size = 100 * 1024
        oss2.resumable_upload(bucket, key, pathname,
                              multipart_threshold=200 * 1024,
                              part_size=part_size,
                              progress_callback=progress_callback,
                              num_threads=1)
        self.assertEqual(stats['previous'], len(content))
        self.assertEqual(stats['ncalled'], oss2.utils.how_many(len(content), part_size) + 1)

        stats = {'previous': -1, 'ncalled': 0}
        oss2.resumable_upload(bucket, key, pathname,
                              multipart_threshold=len(content) + 100,
                              progress_callback=progress_callback)
        self.assertEqual(stats['previous'], len(content))

        bucket.delete_object(key)

    def _rebuild_record(self, filename, store, bucket, key, upload_id, part_size=None, upload_context=None):
        abspath = os.path.abspath(filename)
        mtime = os.path.getmtime(filename)
        size = os.path.getsize(filename)

        record = {'op_type': 'ResumableUpload', 'upload_id': upload_id, 'file_path': abspath, 'size': size,
                  'mtime': mtime, 'bucket': bucket.bucket_name, 'key': key, 'part_size': part_size}

        if upload_context:
            material = upload_context.content_crypto_material
            material_record = {'wrap_alg': material.wrap_alg, 'cek_alg': material.cek_alg,
                               'encrypted_key': utils.b64encode_as_string(material.encrypted_key),
                               'encrypted_iv': utils.b64encode_as_string(material.encrypted_iv),
                               'mat_desc': material.mat_desc}
            record['content_crypto_material'] = material_record

        store_key = store.make_store_key(bucket.bucket_name, key, abspath)
        store.put(store_key, record)

    def __test_resume(self, content_size, uploaded_parts, expected_unfinished=0):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        part_size = 100 * 1024
        num_parts = (content_size + part_size - 1) // part_size

        key = 'resume-' + random_string(32)
        content = random_bytes(content_size)
        encryption_flag = isinstance(bucket, oss2.CryptoBucket)

        context = None
        pathname = self._prepare_temp_file(content)
        if encryption_flag:
            context = models.MultipartUploadCryptoContext(content_size, part_size)
            upload_id = bucket.init_multipart_upload(key, upload_context=context).upload_id
        else:
            upload_id = bucket.init_multipart_upload(key).upload_id

        for part_number in uploaded_parts:
            start = (part_number - 1) * part_size
            if part_number == num_parts:
                end = content_size
            else:
                end = start + part_size

            if encryption_flag:
                bucket.upload_part(key, upload_id, part_number, content[start:end], upload_context=context)
            else:
                bucket.upload_part(key, upload_id, part_number, content[start:end])

        self._rebuild_record(pathname, oss2.resumable.make_upload_store(), bucket, key, upload_id, part_size, context)
        oss2.resumable_upload(bucket, key, pathname, multipart_threshold=0, part_size=100 * 1024)

        result = bucket.get_object(key)
        self.assertEqual(content, result.read())

        self.assertEqual(len(list(oss2.ObjectUploadIterator(self.bucket, key))), expected_unfinished)

        bucket.delete_object(key)

    def test_resume_empty(self):
        self.__test_resume(250 * 1024, [])

    def test_resume_continuous(self):
        self.__test_resume(500 * 1024, [1, 2])

    def test_resume_hole_mid(self):
        self.__test_resume(510 * 1024, [1, 4])

    def test_resume_hole_end(self):
        self.__test_resume(300 * 1024 + 1, [4])

    def __test_interrupt(self, content_size, failed_part_number, expected_unfinished=0, modify_record_func=None):
        bucket = random.choice([self.bucket, self.rsa_crypto_bucket, self.kms_crypto_bucket])
        encryption_flag = isinstance(bucket, oss2.CryptoBucket)
        orig_upload_part = bucket.upload_part

        if encryption_flag:
            def upload_part(self, key, upload_id, part_number, data, progress_callback=None, headers=None,
                            upload_context=None):
                if part_number == failed_part_number:
                    raise RuntimeError
                else:
                    return orig_upload_part(key, upload_id, part_number, data, progress_callback, headers,
                                            upload_context)
        else:
            def upload_part(self, key, upload_id, part_number, data, progress_callback=None, headers=None):
                if part_number == failed_part_number:
                    raise RuntimeError
                else:
                    return orig_upload_part(key, upload_id, part_number, data, progress_callback, headers)

        key = 'ResumableUpload-' + random_string(32)
        content = random_bytes(content_size)

        pathname = self._prepare_temp_file(content)

        if encryption_flag:
            with patch.object(oss2.CryptoBucket, 'upload_part', side_effect=upload_part,
                              autospec=True) as mock_upload_part:
                self.assertRaises(RuntimeError, oss2.resumable_upload, bucket, key, pathname, multipart_threshold=0,
                                  part_size=100 * 1024)
        else:
            with patch.object(oss2.Bucket, 'upload_part', side_effect=upload_part, autospec=True) as mock_upload_part:
                self.assertRaises(RuntimeError, oss2.resumable_upload, bucket, key, pathname, multipart_threshold=0,
                                  part_size=100 * 1024)

        if modify_record_func:
            modify_record_func(oss2.resumable.make_upload_store(), bucket.bucket_name, key, pathname)

        oss2.resumable_upload(bucket, key, pathname, multipart_threshold=0, part_size=100 * 1024)

        self.assertEqual(len(list(oss2.ObjectUploadIterator(self.bucket, key))), expected_unfinished)

    def test_interrupt_at_start(self):
        self.__test_interrupt(310 * 1024, 1)

    def test_interrupt_at_mid(self):
        self.__test_interrupt(510 * 1024, 3)

    def test_interrupt_at_end(self):
        self.__test_interrupt(500 * 1024 - 1, 5)

    def test_record_invalid_op(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('op_type', 'ResumableDownload'),
                              expected_unfinished=1)

    def test_record_invalid_upload_id(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('upload_id', random.randint(1, 100)),
                              expected_unfinished=1)

    def test_record_invalid_file_path(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('file_path', random.randint(1, 100)),
                              expected_unfinished=1)

    def test_record_invalid_size(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('size', 'invalid_size_type'),
                              expected_unfinished=1)

    def test_record_invalid_mtime(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('mtime', 'invalid_mtime'),
                              expected_unfinished=1)

    def test_record_invalid_bucket(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('bucket', random.randint(1, 100)),
                              expected_unfinished=1)

    def test_record_invalid_key(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('key', random.randint(1, 100)),
                              expected_unfinished=1)

    def test_record_invalid_part_size(self):
        self.__test_interrupt(500 * 1024, 1,
                              modify_record_func=self.__make_corrupt_record('part_size', 'invalid_part_size'),
                              expected_unfinished=1)

    def test_file_changed_mtime(self):
        def change_mtime(store, bucket_name, key, pathname):
            time.sleep(2)
            os.utime(pathname, (time.time(), time.time()))

        self.__test_interrupt(500 * 1024, 3,
                              modify_record_func=change_mtime,
                              expected_unfinished=1)

    def test_file_changed_size(self):
        def change_size(store, bucket_name, key, pathname):
            mtime = os.path.getmtime(pathname)

            with open(pathname, 'wb') as f:
                f.write(random_bytes(500 * 1024 - 1))

            os.utime(pathname, (mtime, mtime))

        self.__test_interrupt(500 * 1024, 3, modify_record_func=change_size, expected_unfinished=1)

    def __make_corrupt_record(self, name, value):
        def corrupt_record(store, bucket_name, key, pathname):
            store_key = store.make_store_key(bucket_name, key, pathname)

            record = store.get(store_key)
            record[name] = value
            store.put(store_key, record)

        return corrupt_record

    def test_upload_large_with_tagging(self):
        
        from oss2.compat import urlquote

        key = random_string(16)
        content = random_bytes(5 * 100 * 1024)

        pathname = self._prepare_temp_file(content)

        headers = dict()
        tagging_header = oss2.headers.OSS_OBJECT_TAGGING

        key1 = 128*'a'
        value1 = 256*'b'

        key2 = '+-:/'
        value2 = ':+:'

        key3 = '中文'
        value3 = '++中文++'

        tag_str = key1 + '=' + value1
        tag_str += '&' + urlquote(key2) + '=' + urlquote(value2)
        tag_str += '&' + urlquote(key3) + '=' + urlquote(value3)

        headers[tagging_header] = tag_str

        result = oss2.resumable_upload(self.bucket, key, pathname, multipart_threshold=200 * 1024, num_threads=3, headers=headers)

        self.assertTrue(result is not None)
        self.assertTrue(result.etag is not None)
        self.assertTrue(result.request_id is not None)

        result = self.bucket.get_object(key)
        self.assertEqual(content, result.read())
        self.assertEqual(result.headers['x-oss-object-type'], 'Multipart')

        result = self.bucket.get_object_tagging(key)
        
        self.assertEqual(3, result.tag_set.len())
        tagging_rule = result.tag_set.tagging_rule
        self.assertEqual(256*'b', tagging_rule[128*'a'])
        self.assertEqual(':+:', tagging_rule['+-:/'])
        self.assertEqual('++中文++', tagging_rule['中文'])

        self.bucket.delete_object_tagging(key)
        
        result = self.bucket.get_object_tagging(key)
        
        self.assertEqual(0, result.tag_set.len())
        self.bucket.delete_object(key)

    def test_upload_sequenial(self):
        endpoint = "http://oss-cn-shanghai.aliyuncs.com"
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-upload-sequential"
        bucket = oss2.Bucket(auth, endpoint, bucket_name)
        bucket.create_bucket()

        key = "test-upload_sequential"
        content = random_bytes(10 * 100 * 1024)
        pathname = self._prepare_temp_file(content)

        oss2.resumable_upload(bucket, key, pathname, multipart_threshold=200 * 1024, part_size=100 * 1024, num_threads=6)
        result = bucket.get_object(key)
        self.assertIsNone(result.resp.headers.get('Content-MD5'))

        # single thread in sequential mode.
        params = {'sequential': ''}
        oss2.resumable_upload(bucket, key, pathname, multipart_threshold=200 * 1024, part_size=100 * 1024, num_threads=1,
                              params=params)
        result = bucket.get_object(key)
        self.assertIsNotNone(result.resp.headers.get('Content-MD5'))

        # sequential mode is confused with multi thread.
        try:
            params = {'sequential': ''}
            oss2.resumable_upload(bucket, key, pathname, multipart_threshold=200 * 1024, part_size=100 * 1024, num_threads=6,
                              params=params)
            self.assertFalse(True, 'should be failed here.')
        except oss2.exceptions.PartNotSequential as e:
            pass

    def test_upload_with_acl(self):
        key = "test-upload-with-acl"
        content = random_bytes(5 * 100 * 1024)
        pathname = self._prepare_temp_file(content)

        # Resumable upload without acl setting, returned result should be "default".
        oss2.resumable_upload(self.bucket, key, pathname, multipart_threshold=2 * 100 * 1024,
                              part_size=100 * 1024, num_threads=6)
        result = self.bucket.get_object_acl(key)
        self.assertEqual(oss2.OBJECT_ACL_DEFAULT, result.acl)

        # Resumable upload with "private" acl setting, returned result should be "private".
        headers = {}
        headers["x-oss-object-acl"] = oss2.OBJECT_ACL_PRIVATE
        oss2.resumable_upload(self.bucket, key, pathname, multipart_threshold=2 * 100 * 1024,
                              part_size=100 * 1024, num_threads=6, headers=headers)
        result = self.bucket.get_object_acl(key)
        self.assertEqual(oss2.OBJECT_ACL_PRIVATE, result.acl)

        # Test the file's size smaller than the threshold.
        headers = {}
        headers["x-oss-object-acl"] = oss2.OBJECT_ACL_PRIVATE
        oss2.resumable_upload(self.bucket, key, pathname, multipart_threshold=10 * 100 * 1024,
                              part_size=100 * 1024, num_threads=6, headers=headers)
        result = self.bucket.get_object_acl(key)
        self.assertEqual(oss2.OBJECT_ACL_PRIVATE, result.acl)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .common import *


class TestInit(OssTestCase):

    def test_set_logger(self):
        oss2.set_stream_logger('oss2', logging.DEBUG)
        self.assertTrue(oss2.logger.name, 'oss2')
        self.assertTrue(oss2.logger.level, logging.DEBUG)

        log_file_path = self.random_filename()
        oss2.set_file_logger(log_file_path, 'oss2', logging.INFO)
        self.assertTrue(oss2.logger.name, 'oss2')
        self.assertTrue(oss2.logger.level, logging.INFO)
        oss2.logger.info("hello, oss2")

        with open(log_file_path,'rb') as f:
            self.assertTrue("hello, oss2" in oss2.to_string(f.read()))

        oss2.set_stream_logger('oss2', logging.CRITICAL)
        oss2.set_file_logger(log_file_path, 'oss2', logging.CRITICAL)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

from .common import *
from oss2.models import ServerSideEncryptionRule, PartInfo
from oss2 import (SERVER_SIDE_ENCRYPTION_KMS, SERVER_SIDE_ENCRYPTION_AES256,
                  SERVER_SIDE_ENCRYPTION_SM4, KMS_DATA_ENCRYPTION_SM4)
from oss2.headers import (OSS_SERVER_SIDE_ENCRYPTION, OSS_SERVER_SIDE_ENCRYPTION_KEY_ID,
                          OSS_SERVER_SIDE_DATA_ENCRYPTION)
from oss2 import determine_part_size, SizedFileAdapter


class TestSSEDataEncryption(OssTestCase):
    def setUp(self):
        OssTestCase.setUp(self)
        self.endpoint = OSS_ENDPOINT

    def test_put_bucket_encryption(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-put-bucket-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        # set SM4
        rule = ServerSideEncryptionRule()
        rule.sse_algorithm = oss2.SERVER_SIDE_ENCRYPTION_SM4
        bucket.put_bucket_encryption(rule)

        result = bucket.get_bucket_encryption()
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, result.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertIsNone(result.kms_data_encryption)

        bucket_info = bucket.get_bucket_info()
        rule = bucket_info.bucket_encryption_rule
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, rule.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertIsNone(result.kms_data_encryption)

        # set KMS and data SM4, and none kms_key_id.
        rule = ServerSideEncryptionRule()
        rule.sse_algorithm = SERVER_SIDE_ENCRYPTION_KMS
        rule.kms_data_encryption = KMS_DATA_ENCRYPTION_SM4
        bucket.put_bucket_encryption(rule)

        result = bucket.get_bucket_encryption()
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, result.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, result.kms_data_encryption)

        bucket_info = bucket.get_bucket_info()
        rule = bucket_info.bucket_encryption_rule
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, rule.sse_algorithm)
        self.assertIsNone(rule.kms_master_keyid)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, rule.kms_data_encryption)

        # set KMS and SM4, and has kms key id
        rule = ServerSideEncryptionRule()
        rule.sse_algorithm = SERVER_SIDE_ENCRYPTION_KMS
        rule.kms_master_keyid = '123'
        rule.kms_data_encryption = KMS_DATA_ENCRYPTION_SM4
        bucket.put_bucket_encryption(rule)

        result = bucket.get_bucket_encryption()
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, result.sse_algorithm)
        self.assertEqual('123', result.kms_master_keyid)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, result.kms_data_encryption)

        bucket_info = bucket.get_bucket_info()
        rule = bucket_info.bucket_encryption_rule
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, rule.sse_algorithm)
        self.assertEqual('123', rule.kms_master_keyid)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, rule.kms_data_encryption)

        # set AES256 and data encryption is not none
        rule = ServerSideEncryptionRule()
        rule.sse_algorithm = SERVER_SIDE_ENCRYPTION_AES256
        rule.kms_data_encryption = KMS_DATA_ENCRYPTION_SM4
        bucket.put_bucket_encryption(rule)

        result = bucket.get_bucket_encryption()
        self.assertEqual(SERVER_SIDE_ENCRYPTION_AES256, result.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertIsNone(result.kms_data_encryption)

        bucket_info = bucket.get_bucket_info()
        rule = bucket_info.bucket_encryption_rule
        self.assertEqual(SERVER_SIDE_ENCRYPTION_AES256, rule.sse_algorithm)
        self.assertIsNone(rule.kms_master_keyid)
        self.assertIsNone(rule.kms_data_encryption)

        # set SM4 and data encryption is not none
        rule = ServerSideEncryptionRule()
        rule.sse_algorithm = SERVER_SIDE_ENCRYPTION_SM4
        rule.kms_data_encryption = KMS_DATA_ENCRYPTION_SM4
        bucket.put_bucket_encryption(rule)

        result = bucket.get_bucket_encryption()
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, result.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertIsNone(result.kms_data_encryption)

        bucket_info = bucket.get_bucket_info()
        rule = bucket_info.bucket_encryption_rule
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, rule.sse_algorithm)
        self.assertIsNone(result.kms_master_keyid)
        self.assertIsNone(result.kms_data_encryption)

    def inner_put_object_with_encryption(self, bucket, object_name, data, sse_algorithm, data_algorithm):
        headers = dict()
        if sse_algorithm:
            headers[OSS_SERVER_SIDE_ENCRYPTION] = sse_algorithm
        if data_algorithm:
            headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = data_algorithm

        result = bucket.put_object(object_name, data, headers=headers)

        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        return ret_sse_algo, ret_data_algo, ret_kms_key_id


    def test_put_object_with_encryption(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-put-object-data-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        object_name = 'test-put-object-none-encryption'
        data = b'a'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, None, None)
        self.assertIsNone(sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

        object_name = 'test-put-object-kms'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_KMS, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNotNone(key_id)

        object_name = 'test-put-object-aes'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_AES256, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_AES256, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

        object_name = 'test-put-object-sm4'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_SM4, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

        object_name = 'test-put-object-kms-sm4'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_KMS, KMS_DATA_ENCRYPTION_SM4)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, data_algo)
        self.assertIsNotNone(key_id)

        object_name = 'test-put-object-aes256-sm4'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_AES256, KMS_DATA_ENCRYPTION_SM4)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_AES256, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

        object_name = 'test-put-object-sm4-sm4'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, SERVER_SIDE_ENCRYPTION_SM4, KMS_DATA_ENCRYPTION_SM4)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_SM4, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

        object_name = 'test-put-object-none-sm4'
        sse_algo, data_algo, key_id = self.inner_put_object_with_encryption(bucket, object_name, data, None, KMS_DATA_ENCRYPTION_SM4)
        self.assertIsNone(sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(key_id)

    def inner_append_object_with_encryption(self, bucket, object_name, position, data, sse_algorithm, data_algorithm):
        headers = dict()
        if sse_algorithm:
            headers[OSS_SERVER_SIDE_ENCRYPTION] = sse_algorithm
        if data_algorithm:
            headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = data_algorithm

        result = bucket.append_object(object_name, position, data, headers=headers)
        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        return ret_sse_algo, ret_data_algo, ret_kms_key_id

    def test_append_object_with_encryption(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-append-object-data-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        # first append
        object_name = 'test-append'
        sse_algo, data_algo, key_id = self.inner_append_object_with_encryption(bucket, object_name, 0, '123', SERVER_SIDE_ENCRYPTION_KMS, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNotNone(key_id)
        kms_key_id = key_id

        # second append with none algorithm
        sse_algo, data_algo, key_id = self.inner_append_object_with_encryption(bucket, object_name, 3, '456', None, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNotNone(key_id)
        self.assertEqual(kms_key_id, key_id)

        # third append with other algorithm
        sse_algo, data_algo, key_id = self.inner_append_object_with_encryption(bucket, object_name, 6, '789', SERVER_SIDE_ENCRYPTION_AES256, None)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNotNone(key_id)
        self.assertEqual(kms_key_id, key_id)

    def test_multipart_upload(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-multipart-upload-data-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        key = 'data-encryption-test-upload-part-object'
        filename = self._prepare_temp_file_with_size(1024 * 1024)

        headers = dict()
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_KMS
        headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = KMS_DATA_ENCRYPTION_SM4

        total_size = os.path.getsize(filename)
        # Set part size
        part_size = determine_part_size(total_size, preferred_size=(100*1024))

        # Init multipart with encryption headers.
        result = bucket.init_multipart_upload(key, headers=headers)
        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, ret_sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, ret_data_algo)
        self.assertIsNotNone(ret_kms_key_id)

        kms_key_id = ret_kms_key_id
        upload_id = result.upload_id
        parts = []

        # Upload part with the encyrption headers will be failed.
        headers = dict()
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_KMS
        headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = KMS_DATA_ENCRYPTION_SM4
        with open(filename, 'rb') as fileobj:
            part_number = 1
            num_to_upload = part_size
            self.assertRaises(oss2.exceptions.InvalidArgument, bucket.upload_part, key, upload_id, part_number,
                                            SizedFileAdapter(fileobj, num_to_upload), headers=headers)

        # Upload part with none encryption headers.
        with open(filename, 'rb') as fileobj:
            part_number = 1
            offset = 0
            while offset < total_size:
                num_to_upload = min(part_size, total_size - offset)

                result = bucket.upload_part(key, upload_id, part_number,
                                            SizedFileAdapter(fileobj, num_to_upload))

                parts.append(PartInfo(part_number, result.etag))

                offset += num_to_upload
                part_number += 1

                ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
                ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
                ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
                self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, ret_sse_algo)
                self.assertEqual(KMS_DATA_ENCRYPTION_SM4, ret_data_algo)
                self.assertEqual(kms_key_id, ret_kms_key_id)

        # Complete multipart upload with encryption headers.
        self.assertRaises(oss2.exceptions.InvalidArgument, bucket.complete_multipart_upload, key, upload_id, parts, headers=headers)

        # Complete multipart upload with none encryption headers.
        result = bucket.complete_multipart_upload(key, upload_id, parts)
        self.assertEqual(result.status, 200)
        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, ret_sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, ret_data_algo)
        self.assertEqual(kms_key_id, ret_kms_key_id)

        bucket.delete_object(key)


    def test_resumable_uoload(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-resumable-upload-data-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        small_object = 'requestpayment-test-resumable-upload-small-object'
        big_object = 'requestpayment-test-resumable-upload-big-object'

        # Create tmp file smaller than multipart_threshold
        file_name = self._prepare_temp_file_with_size(150 * 1024)

        # Resumale upload small object
        headers = dict()
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_KMS
        headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = KMS_DATA_ENCRYPTION_SM4
        result = oss2.resumable_upload(bucket, small_object, file_name,
                                       multipart_threshold=(200 * 1024), num_threads=2, part_size=(100 * 1024),
                                       headers=headers)
        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, ret_sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, ret_data_algo)
        self.assertIsNotNone(ret_kms_key_id)
        self.assertEqual(result.status, 200)
        bucket.delete_object(small_object)

        # Start big file test
        # Create big file bigger than multipart_threshold
        file_name = self._prepare_temp_file_with_size(11 * 1024 * 1024)

        headers = dict()
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_KMS
        headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = KMS_DATA_ENCRYPTION_SM4
        result = oss2.resumable_upload(bucket, big_object, file_name,
                                       multipart_threshold=(200 * 1024), num_threads=2, part_size=(100 * 1024),
                                       headers=headers)

        self.assertEqual(result.status, 200)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, ret_sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, ret_data_algo)
        self.assertIsNotNone(ret_kms_key_id)
        self.assertEqual(result.status, 200)
        bucket.delete_object(big_object)

    def test_copy_object(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-copy-object-data-encryption"
        bucket = oss2.Bucket(auth, self.endpoint, bucket_name)
        bucket.create_bucket()

        object_name = 'test-copy-object-src'
        data = b'a' * 1024
        user_header_key = 'x-oss-meta-user1'
        user_header_value = 'user_value'

        headers = dict()
        headers[user_header_key] = user_header_value
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_KMS
        headers[OSS_SERVER_SIDE_DATA_ENCRYPTION] = KMS_DATA_ENCRYPTION_SM4

        result = bucket.put_object(object_name, data, headers=headers)
        sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, data_algo)
        self.assertIsNotNone(kms_key_id)

        result = bucket.head_object(object_name)
        sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        ret_value = result.headers.get(user_header_key)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_KMS, sse_algo)
        self.assertEqual(KMS_DATA_ENCRYPTION_SM4, data_algo)
        self.assertIsNotNone(kms_key_id)
        self.assertEqual(user_header_value, ret_value)

        #  test normal copy objects
        dest_name = 'test-copy-object-dest'
        bucket.copy_object(bucket_name, object_name, dest_name)

        result = bucket.head_object(dest_name)
        ret_sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        ret_data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        ret_kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertIsNone(ret_sse_algo)
        self.assertIsNone(ret_data_algo)
        self.assertIsNone(ret_kms_key_id)

        # test copy object with specified encryption headers
        headers = dict()
        headers[user_header_key] = user_header_value
        headers[OSS_SERVER_SIDE_ENCRYPTION] = SERVER_SIDE_ENCRYPTION_AES256
        bucket.copy_object(bucket_name, object_name, dest_name, headers=headers)

        result = bucket.head_object(dest_name)
        sse_algo = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION)
        data_algo = result.headers.get(OSS_SERVER_SIDE_DATA_ENCRYPTION)
        kms_key_id = result.headers.get(OSS_SERVER_SIDE_ENCRYPTION_KEY_ID)
        self.assertEqual(SERVER_SIDE_ENCRYPTION_AES256, sse_algo)
        self.assertIsNone(data_algo)
        self.assertIsNone(kms_key_id)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import json
import base64
import os

import oss2


# 以下代码展示了上传回调的用法。

# put_object/complete_multipart_upload支持上传回调，resumable_upload不支持。
# 回调服务器(callbacke server)的示例代码请参考 http://shinenuaa.oss-cn-hangzhou.aliyuncs.com/images/callback_app_server.py.zip
# 您也可以使用OSS提供的回调服务器 http://oss-demo.aliyuncs.com:23450，调试您的程序。调试完成后换成您的回调服务器。

# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
access_key_id = os.getenv('OSS_TEST_ACCESS_KEY_ID', '<你的AccessKeyId>')
access_key_secret = os.getenv('OSS_TEST_ACCESS_KEY_SECRET', '<你的AccessKeySecret>')
bucket_name = os.getenv('OSS_TEST_BUCKET', '<你的Bucket>')
endpoint = os.getenv('OSS_TEST_ENDPOINT', '<你的访问域名>')


# 确认上面的参数都填写正确了
for param in (access_key_id, access_key_secret, bucket_name, endpoint):
    assert '<' not in param, '请设置参数：' + param

key = 'quote.txt'
content = "Anything you're good at contributes to happiness."

# 创建Bucket对象，所有Object相关的接口都可以通过Bucket对象来进行
bucket = oss2.Bucket(oss2.Auth(access_key_id, access_key_secret), endpoint, bucket_name)

"""
put_object上传回调
"""

# 准备回调参数，更详细的信息请参考 https://help.aliyun.com/document_detail/31989.html
callback_dict = {}
callback_dict['callbackUrl'] = 'http://oss-demo.aliyuncs.com:23450'
callback_dict['callbackHost'] = 'oss-cn-hangzhou.aliyuncs.com'
callback_dict['callbackBody'] = 'filename=${object}&size=${size}&mimeType=${mimeType}'
callback_dict['callbackBodyType'] = 'application/x-www-form-urlencoded'
# 回调参数是json格式，并且base64编码
callback_param = json.dumps(callback_dict).strip()
base64_callback_body = oss2.utils.b64encode_as_string(callback_param)
# 回调参数编码后放在header中传给oss
headers = {'x-oss-callback': base64_callback_body}

# 上传并回调
result = bucket.put_object(key, content, headers)

# 上传并回调成功status为200，上传成功回调失败status为203
assert result.status == 200
# result.resp的内容为回调服务器返回的内容
assert result.resp.read() == b'{"Status":"OK"}'

# 确认文件上传成功
result = bucket.head_object(key)
assert result.headers['x-oss-hash-crc64ecma'] == '108247482078852440'

# 删除上传的文件
bucket.delete_object(key)

"""
分片上传回调
"""

# 分片上传回调
# 初始化上传任务
parts = []
upload_id = bucket.init_multipart_upload(key).upload_id
# 上传分片
result = bucket.upload_part(key, upload_id, 1, content)
parts.append(oss2.models.PartInfo(1, result.etag, size = len(content), part_crc = result.crc))
# 完成上传并回调
result = bucket.complete_multipart_upload(key, upload_id, parts, headers)

# 上传并回调成功status为200，上传成功回调失败status为203
assert result.status == 200
# result.resp的内容为回调服务器返回的内容
assert result.resp.read() == b'{"Status":"OK"}'

# 确认文件上传成功
result = bucket.head_object(key)
assert result.headers['x-oss-hash-crc64ecma'] == '108247482078852440'

# 删除上传的文件
bucket.delete_object(key)
